<?php

namespace Drupal\commerce_cart_estimate;

use CommerceGuys\Addressing\Subdivision\SubdivisionRepositoryInterface;
use Drupal\commerce_cart_estimate\Event\CartEstimateEvents;
use Drupal\commerce_cart_estimate\Event\SelectShippingRateEvent;
use Drupal\commerce_cart_estimate\Exception\CartEstimateException;
use Drupal\commerce_cart_estimate\Exception\NoApplicableShippingRatesException;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_order\OrderRefreshInterface;
use Drupal\commerce_shipping\ShipmentManagerInterface;
use Drupal\commerce_shipping\ShippingOrderManagerInterface;
use Drupal\profile\Entity\ProfileInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

/**
 * Provides a service for estimating taxes/shipping rates for orders.
 */
class Estimator implements EstimatorInterface {

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * The order refresh service.
   *
   * @var \Drupal\commerce_order\OrderRefreshInterface
   */
  protected $orderRefresh;

  /**
   * The shipping order manager.
   *
   * @var \Drupal\commerce_shipping\ShippingOrderManagerInterface
   */
  protected $shippingOrderManager;

  /**
   * The shipment manager.
   *
   * @var \Drupal\commerce_shipping\ShipmentManagerInterface
   */
  protected $shipmentManager;

  /**
   * The subdivision repository.
   *
   * @var \CommerceGuys\Addressing\Subdivision\SubdivisionRepositoryInterface
   */
  protected $subdivisionRepository;

  /**
   * Constructs a new Estimator object.
   *
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher.
   * @param \Drupal\commerce_order\OrderRefreshInterface $order_refresh
   *   The order refresh service.
   * @param \Drupal\commerce_shipping\ShippingOrderManagerInterface $shipping_order_manager
   *   The shipping order manager.
   * @param \Drupal\commerce_shipping\ShipmentManagerInterface $shipment_manager
   *   The shipment manager.
   * @param \CommerceGuys\Addressing\Subdivision\SubdivisionRepositoryInterface $subdivision_repository
   *   The subdivision repository.
   */
  public function __construct(EventDispatcherInterface $event_dispatcher, OrderRefreshInterface $order_refresh, ShippingOrderManagerInterface $shipping_order_manager, ShipmentManagerInterface $shipment_manager, SubdivisionRepositoryInterface $subdivision_repository) {
    $this->eventDispatcher = $event_dispatcher;
    $this->orderRefresh = $order_refresh;
    $this->shippingOrderManager = $shipping_order_manager;
    $this->shipmentManager = $shipment_manager;
    $this->subdivisionRepository = $subdivision_repository;
  }

  /**
   * {@inheritdoc}
   */
  public function buildShippingProfile(OrderInterface $order, array $address) {
    // Attempt to automatically figure out the state for "US" addresses,
    // for more accurate shipping estimates.
    if ($address['country_code'] === 'US' && !empty($address['postal_code']) && empty($address['administrative_area'])) {
      $subdivision_repository = \Drupal::service('address.subdivision_repository');
      $subdivisions = $subdivision_repository->getAll(['US']);
      $postal_code = $address['postal_code'];
      // Attempt to match the subdivision.
      foreach ($subdivisions as $subdivision) {
        preg_match('/' . $subdivision->getPostalCodePattern() . '/i', $postal_code, $matches);
        if (!isset($matches[0]) || strpos($postal_code, $matches[0]) !== 0) {
          continue;
        }
        $address['administrative_area'] = $subdivision->getCode();
        break;
      }
    }

    return $this->shippingOrderManager->createProfile($order, [
      'address' => $address,
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function estimate(OrderInterface $order, ProfileInterface $profile) {
    if (!$this->shippingOrderManager->isShippable($order)) {
      throw new \InvalidArgumentException('The provided order is not shippable.');
    }
    $rated_order = clone $order;
    // Clear the existing shipments reference, to ensure existing shipments
    // are not removed by the shipping order manager.
    $rated_order->set('shipments', NULL);
    $shipments = $this->shippingOrderManager->pack($rated_order, $profile);

    // @todo: throw an exception when this happens?
    if (!$shipments) {
      /** @var \Drupal\address\Plugin\Field\FieldType\AddressItem $address */
      $address = $profile->get('address')->first();
      throw new CartEstimateException(sprintf('Could not pack the order when estimating shipping. (Country code: %s, Postal code: %s).', $address->getCountryCode(), $address->getPostalCode()));
    }
    $rated_order->set('shipments', $shipments);

    $rates = [];
    // Rate the shipments returned.
    foreach ($shipments as $shipment) {
      $shipment->order_id->entity = $rated_order;
      $rates = $this->shipmentManager->calculateRates($shipment);

      // When no rates could be calculated, throw an exception.
      if (!$rates) {
        /** @var \Drupal\address\Plugin\Field\FieldType\AddressItem $address */
        $address = $profile->get('address')->first();
        throw new NoApplicableShippingRatesException(sprintf('No applicable shipping rates for the following partial address: (Country code: %s, Postal code: %s).', $address->getCountryCode(), $address->getPostalCode()));
      }
      // Default to the first rate from the rates array.
      $rate_to_apply = reset($rates);
      // Allow customizing the shipping rate that is going to be applied for
      // the cart estimate via an event.
      $event = new SelectShippingRateEvent($rates, $rate_to_apply, $shipment);
      $this->eventDispatcher->dispatch(CartEstimateEvents::SELECT_SHIPPING_RATE, $event);
      $this->shipmentManager->applyRate($shipment, $event->getRate());
    }

    // Refresh the order so that shipping/tax adjustments are applied.
    // Add a custom flag to the order so that custom order processors can target
    // the order refresh performed for estimating shipping.
    $rated_order->setData('commerce_cart_estimate_refresh', TRUE);
    $this->orderRefresh->refresh($rated_order);
    $rated_order->recalculateTotalPrice();

    return new CartEstimateResult($rated_order, $rates);
  }

}
